/*
 * File struct manipulation.
 */
#include <strings.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/inode.h>
#include <sys/socketvar.h>
#include <sys/ioctl.h>
#include <sys/tty.h>		/* For ttyinfo */
#include "pmon.h"
#ifdef NFS
#include <sys/vnode.h>
#endif NFS

extern int errno;
char *malloc();
#define	DTYPE_INODE	1	/* In sys/file.h but ifdef'd on kernel! */
#define	DTYPE_SOCKET	2
#ifdef	ultrix
#define	DTYPE_PORT	3
#define	DTYPE_PIPE	4
#endif	ultrix

/* The root device's major,minor */
dev_t	rootdev;

int	fdbase = 0;

/* File struct flag bits - XXX */
static	char *fileflags[] = {
	"READ",		/* descriptor read/receive'able */
	"WRITE",	/* descriptor write/send'able */
	"NDELAY",	/* no delay */
	"APPEND",	/* append on each write */
	"MARK",		/* mark during gc() */
	"DEFER",	/* defer for next gc pass */
	"ASYNC",	/* signal pgrp when data ready */
	"SHLOCK",	/* shared lock present */
	"EXLOCK",	/* exclusive lock present */
			/* open only modes */
	"CREAT",	/* create if nonexistant */
	"TRUNC",	/* truncate to zero length */
	"EXCL"		/* error if already created */
};

/*
 *	F I L E _ T Y P E
 *
 * Return a string with name for file type code
 */
char *
file_type(ftype)
long ftype;
{
	static char ftbuf[20];
	switch (ftype) {
	case DTYPE_INODE:
		return("inode");
	case DTYPE_SOCKET:
		return("sock");
#ifdef	DTYPE_PORT:
	case DTYPE_PORT:
		return("port");
#endif	DTYPE_PORT:
#ifdef	DTYPE_PIPE:
	case DTYPE_PIPE:
		return("pipe");
#endif	DTYPE_PIPE:
	default:
		sprintf(ftbuf, "%x", ftype);
	}
}

/*
 *	F F L A G S
 *
 * Return a string with names for set file flag bits.
 */
char *
fflags(flags)
unsigned long flags;
{
	int	i;
	static	char	buf[40];

	buf[0] = NULL;
	for( i = 11; i >= 0; i-- ) {
		if( flags & (1<<i) ) {
			strcat( buf, fileflags[i] );
			strcat( buf, " " );
		}
	}

	return( buf );
}

/*
 *	O P E N F I L E S
 *
 * Return a string of 1's and 0's showing
 *   which file descriptors are open.
 */
char *
openfiles()
{
	static char buf[NOFILE+1];
	int	i;

	for( i = 0; i < NOFILE; i++ ) {
		if( u.u_ofile[i] )
			buf[i] = '1';
		else
			buf[i] = '0';
	}

	return (buf);
}

/*
 * We need to read the /dev directory and save all entires
 *  of (dev, inode, filename) found there.
 *
 * Then the (dev,inode) naming strategy becomes:
 *
 *   If it's dev is the root and I've got it's inode number
 *	then display the name of the inode (it's a device).
 *	else display name for dev + inumber (it's a file).
 *
 * Tables are dynamically allocated and hashed by dev and inumber.
 */
typedef	struct	dev_ent	{
	struct	dev_ent	*inext;	/* list linked by inode hash */
	struct	dev_ent	*dnext;	/* list linked by dev hash */
	int	dev;
	int	inode;
	int	type;		/* Character or Block special */
	char	name[1];	/* malloc'd as variable length */
} dev_ent;
/*
 * The hash tables.
 * Perhaps there's a "smart" way to hash
 * these to avoid clustering? - XXX
 */
dev_ent *devinode[256], *devdev[256];
#define	IHASH(x)	((x) & 0xff)
#define	DHASH(x)	((x) & 0xff)

/*
 *	F I N D I N O D E
 *
 * Return pointer to dev entry info
 *  or NULL if inode not in table.
 */
dev_ent *
findinode( type, inum )
int	type, inum;
{
	dev_ent	**dp;

	dp = &devinode[ IHASH(inum) ];
	while( *dp != NULL ) {
		if( (*dp)->inode == inum && (*dp)->type == type )
			return( *dp );
		dp = &((*dp)->inext);
	}

	return( (dev_ent *)NULL );
}

/*
 *	F I N D D E V
 *
 * Return pointer to dev entry info
 *  or NULL if inode not in table.
 * Type is given: IFBLK, IFCHR
 */
dev_ent *
finddev( dev, type )
int	dev;
int	type;
{
	dev_ent	**dp;

	dp = &devdev[ DHASH(dev) ];
	while( *dp != NULL ) {
		if( (*dp)->dev == dev && (*dp)->type == type )
			return( *dp );
		dp = &((*dp)->dnext);
	}

	return( (dev_ent *)NULL );
}

/*
 *	B U I L D _ D E V T A B L E
 *
 * Read /dev and save info on all special files.
 * Link them into hash tables by inode and dev.
 */
build_devtable()
{
	int	i, num, len;
	struct	direct	*((*list)[]);
	struct	stat	stbuf;
	static	char	path[] = "/dev/                         ";
	dev_ent	*dev, **place;

	/* Find the Root dev number */
	if( stat( "/dev", &stbuf ) < 0 ) {
		fprintf( stderr, "stat failed: errno = %d\n", errno );
	}
	rootdev = stbuf.st_dev;

	num = scan_devdir( "/dev/");
#ifdef	ultrix
	/* scan for possible packetfilter devices */
	num += scan_devdir( "/dev/pf/");
#endif	ultrix
/*	fprintf( stderr, "%d entries in devtable\n", num );*/
}

/*
 *	S C A N _ D E V D I R
 *
 * Read a directory and save info on all special files.
 * Link them into hash tables by inode and dev.
 */
scan_devdir(dirname)
char *dirname;
{
	int	dirname_len;
	int	i, num, len;
	struct	direct	*((*list)[]);
	struct	stat	stbuf;
	static	char	path[] = "                              ";
	dev_ent	*dev, **place;

	dirname_len = strlen(dirname);
	strcpy(path, dirname);

	num = scandir( dirname, &list, 0, 0 );
	num = 0;
	/* skip . and .. */
	for( i = 2; ((*list)[i]) != NULL; i++ ) {
		strcpy( &path[dirname_len], ((*list)[i])->d_name );
		stat( path, &stbuf );
		if( stbuf.st_mode & (S_IFCHR | S_IFBLK) ) {
			/*
			 * Save the info
			 */
			len = strlen( ((*list)[i])->d_name ) + 1;
			dev = (dev_ent *)malloc( sizeof(dev_ent) + len );
			dev->dev = stbuf.st_rdev;
			dev->inode = stbuf.st_ino;
			dev->type = stbuf.st_mode & S_IFMT;
			bcopy( ((*list)[i])->d_name, dev->name, len );
			dev->inext = dev->dnext = NULL;
			/*
			 * Link it into hash tables
			 */
			place = &devinode[ IHASH( dev->inode ) ];
			while( *place != NULL ) place = &((*place)->inext);
			*place = dev;
			place = &devdev[ DHASH( dev->dev ) ];
			while( *place != NULL ) place = &((*place)->dnext);
			*place = dev;
#ifdef DEBUG
			fprintf( stderr, "%s\n", devinode[IHASH(dev->inode)]->name );
			fprintf( stderr, "  dev = %d\n", devinode[IHASH(dev->inode)]->dev );
#endif DEBUG
			num++;
		}
	}
	return(num);
}

/*
 *	D I S P F I L E S
 *
 * Display open file information
 *  at screen position bx, by.
 */
dispfiles( by, bx )
int bx, by;
{
	int	i, num_disp, num_seen, type;
	struct	file	file;
#ifdef	ultrix
	struct	gnode	gnode;
#else
	struct	inode	inode;
#endif	ultrix
	dev_ent	*dp;
#ifdef NFS
	struct	vnode	vnode;
#endif NFS

	num_disp = num_seen = 0;
	for( i = 0; i < NOFILE && num_disp < FDISPLAY; i++ ) {
		if( u.u_ofile[i] == 0 )
			continue;

		/* See if this one's on the screen */
		if( num_seen < fdbase ) {
			num_seen++;
			continue;
		}

		/* Get file struct info */
		lseek( kmem, u.u_ofile[i], 0 );
		if( read( kmem, &file, sizeof(file) ) < sizeof(file) ) {
			fprintf( stderr, "pmon: Error reading file struct\n" );
			return;
		}
		mvprintw( by+num_disp, bx, "%2d %5s %2d 0x%-10x %s\n",
			i,
			file_type(file.f_type),
			file.f_count,
			file.f_offset,
			fflags(file.f_flag) );

		switch (file.f_type) {
		case DTYPE_INODE:
			break;	/* drop through to code below */
		case DTYPE_SOCKET:
			sockinfo( by+num_disp, bx+50, file.f_data );
			num_disp++;	/* this is poor */
			continue;
#ifdef	DTYPE_PORT:
		case DTYPE_PORT:
			num_disp++;	/* this is poor */
			continue;
#endif	DTYPE_PORT:
#ifdef	DTYPE_PIPE:
		case DTYPE_PIPE:
			num_disp++;	/* this is poor */
			continue;
#endif	DTYPE_PIPE:
		default:
			num_disp++;	/* this is poor */
			continue;
		}

		/* code for DTYPE_INODE */

		/* Get inode struct info */
#ifdef NFS
		/* get vnode */
		if( lseek( kmem, file.f_data, 0 ) == -1 ) {
			fprintf( stderr, "pmon: seek to vnode failed!\n" );
		}
		if( read( kmem, &vnode, sizeof(vnode) ) < sizeof(vnode) ) {
			fprintf( stderr, "pmon: Error reading vnode struct\n" );
			return;
		}
		/* then inode */
		if( lseek( kmem, vnode.v_data, 0 ) == -1 ) {
			fprintf( stderr, "pmon: seek to inode failed!\n" );
		}
#else
		if( lseek( kmem, file.f_data, 0 ) == -1 ) {
			fprintf( stderr, "pmon: seek to inode failed!\n" );
		}
#endif NFS
#ifdef	ultrix
		if( read( kmem, &gnode, sizeof(gnode) ) < sizeof(gnode) ) {
			fprintf( stderr, "pmon: Error reading gnode struct\n" );
			return;
		}
		/* File name logic */
		type = gnode.g_mode & IFMT;	/* IFCHR, IFBLK, IFSOCK, etc. */
		if( gnode.g_dev == rootdev && (dp = findinode(type, gnode.g_number)) != NULL ) {
			/* It's a device */
			mvprintw( by+num_disp, bx+50, "%s\n", dp->name );
		} else if( (dp = finddev(gnode.g_dev, IFBLK)) != NULL ) {
			/* File on device */
			mvprintw( by+num_disp, bx+50, "%s: %d\n", dp->name, gnode.g_number );
		} else {
			/* Don't know what it is! */
			mvprintw( by+num_disp, bx+50, "(%5d,%5d) %5d\n",
				major(gnode.g_dev), minor(gnode.g_dev), gnode.g_number );
		}
#else
		if( read( kmem, &inode, sizeof(inode) ) < sizeof(inode) ) {
			fprintf( stderr, "pmon: Error reading inode struct\n" );
			return;
		}
		/* File name logic */
		type = inode.i_mode & IFMT;	/* IFCHR, IFBLK, IFSOCK, etc. */
		if( inode.i_dev == rootdev && (dp = findinode(type, inode.i_number)) != NULL ) {
			/* It's a device */
			mvprintw( by+num_disp, bx+50, "%s\n", dp->name );
		} else if( (dp = finddev(inode.i_dev, IFBLK)) != NULL ) {
			/* File on device */
			mvprintw( by+num_disp, bx+50, "%s: %d\n", dp->name, inode.i_number );
		} else {
			/* Don't know what it is! */
			mvprintw( by+num_disp, bx+50, "(%5d,%5d) %5d\n",
				major(inode.i_dev), minor(inode.i_dev), inode.i_number );
		}
#endif	ultrix
		num_disp++;
		num_seen++;
	}
	while( num_disp < FDISPLAY ) {
		mvprintw( by+num_disp, bx, "\n" );
		num_disp++;
	}
}

/* Socket state bits - XXX */
static
char	*sockstates[] = {
	"NOFDREF",	/* no file table ref any more */
	"CONNECTED",	/* socket connected to a peer */
	"CONNECTING",	/* in process of connecting to peer */
	"DISCONNECTING",/* in process of disconnecting */
	"CANTSENDMORE",	/* can't send more data to peer */
	"CANTRCVMORE",	/* can't receive more data from peer */
	"RCVATMARK",	/* at mark on input */
	"PRIV",		/* privileged for broadcast, raw... */
	"NBIO",		/* non-blocking ops */
	"ASYNC"		/* async i/o notify */
};

/*
 *	S S T A T E S
 * Used by sockinfo to print state string.
 */
static
char *
sstates(state)
unsigned long state;
{
	int	i;
	static	char	buf[80];

	buf[0] = NULL;
	for( i = 9; i >= 0; i-- ) {
		if( state & (1<<i) ) {
			strcat( buf, sockstates[i] );
			strcat( buf, " " );
		}
	}

	return( buf );
}

/*
 *	S O C K I N F O
 *
 * Display socket information at the given screen postion.
 */
static
sockinfo( y, x, addr )
int y, x;
struct socket *addr;
{
	struct	socket	sock;

	/* Get socket struct info */
	lseek( kmem, addr, 0 );
	if( read( kmem, &sock, sizeof(sock) ) < sizeof(sock) ) {
		fprintf( stderr, "pmon: Error reading inode struct\n" );
		return;
	}
	mvprintw( y, x, "%s %d+%di %do\n", sstates(sock.so_state),
		sock.so_rcv.sb_cc, sock.so_rcv.sb_mbcnt,
		sock.so_snd.sb_cc );
}

/*
 *	T T Y I N F O
 *
 * Controlling tty info.
 *  Put here because this is where all the device info was, sigh.
 */
ttyinfo( y, x )
int y, x;
{
	dev_ent	*dp;
	struct	tty	tty;

	/* See if there is a tty struct associated with this proc */
#ifdef	ultrix
	if( pr.p_ttyp == NULL ) {
#else
	if( u.u_ttyp == NULL ) {
#endif	ultrix
		mvprintw( y, x, " tty: %-16s", "[none]" );
		return;
	}

	/* Look up the device name - should we use tty.t_dev? XXX */
	dp = finddev( u.u_ttyd, IFCHR );
	if( dp == NULL ) {
		mvprintw( y, x, " tty: %-16s", "[?]" );
		return;
	}

	/* Get tty struct info */
#ifdef	ultrix
	lseek( kmem, pr.p_ttyp, 0 );
#else
	lseek( kmem, u.u_ttyp, 0 );
#endif	ultrix
	if( read( kmem, &tty, sizeof(tty) ) != sizeof(tty) ) {
		fprintf( stderr, "pmon: Error reading tty struct\n" );
		return;
	}
	mvprintw( y, x, " tty: (%5d) %-8s", tty.t_pgrp, dp->name );
}
